/*
 * Copyright (c) Microsoft Corporation. All rights reserved.
 * Licensed under the MIT License.
 * See License.txt in the project root for license information.
 */

#ifndef __FP_COMMON_H__
#define __FP_COMMON_H__

#define UNUSED(x) (void)(x)
#define LIKELY(expr) __builtin_expect((expr), 1)
#define UNLIKELY(expr) __builtin_expect((expr), 0)

#define IN
#define OUT

#include <atomic>
#include <chrono>
#include <cstdbool>
#include <cstdint>
#include <mutex>
#include <jni.h>

class VscContext
{
private:
    uint64_t hdrMagic;
    std::atomic<uint64_t> status;
    std::atomic<uint64_t> fdOpenCount;
    int32_t vscFd;
    int32_t vscBadFd;
    std::chrono::steady_clock::time_point vscUnavailTimestamp;
    std::mutex vscCtxLock;
    uint64_t ftrMagic;

public:
    VscContext();
    ~VscContext();

    int InitVscDevice();
    void CleanupVscDevice(bool reinit);
    int ReinitVscFd();
    int IssueVscSyncRequest (IN const char *clientRequestId,
        IN uint32_t timeoutMsec,
        IN const char *pInputBuf,
        IN uint32_t inputBufLen,
        OUT char *pDataBuf,
        IN uint32_t dataBufLen,
        IN uint32_t dataValid,
        OUT char *pResponseBuf,
        IN uint32_t responseBufLen,
        OUT int32_t *pLocalErrno,
        OUT uint32_t *pHostResponseLen);
};

class FastpathJniContext {
private:
    std::atomic<uint64_t> inited;
    std::mutex fpJniContextLock;

public:
    std::atomic<bool> testMode;

    FastpathJniContext()
        :inited {0},
         testMode {false}
    {}

    ~FastpathJniContext()
    {
        inited = 0;
        testMode = false;
    }

    void InitFastpathContext(jboolean inTestMode)
    {
        std::lock_guard<std::mutex> lk_guard {fpJniContextLock};

        if (inited == 0) {
            testMode = inTestMode;
            inited = 1;
        }
    }
};

static constexpr uint64_t SHAREDLIB_CTX_HDR_MAGIC = 0x0B0B0B0B0BADCAFE;
static constexpr uint64_t SHAREDLIB_CTX_FTR_MAGIC = 0x0D0D0D0D0BADCAFE;
class SharedLibContext {
public:
    uint64_t hdrMagic;
    VscContext vscContext;
    FastpathJniContext fastpathJniContext;
    uint64_t ftrMagic;

    SharedLibContext()
        :hdrMagic {SHAREDLIB_CTX_HDR_MAGIC},
         ftrMagic {SHAREDLIB_CTX_FTR_MAGIC}
    {
        vscContext.InitVscDevice();
    }

    ~SharedLibContext()
    {
    }
};

#endif // __FP_COMMON_H__
