/*
 * Copyright (c) Microsoft Corporation. All rights reserved.
 * Licensed under the MIT License.
 * See License.txt in the project root for license information.
 */

#include <cinttypes>
#include <cerrno>

#include <cstring>
#include <map>

#include <com_azure_storage_fastpath_driver_FastpathDriver.h>

#include <vsc/fastpath.h>
#include <util/crc64.h>
#include "exceptions.h"
#include "test/fake_responses.h"

/*
 * From FastpathRequestHeader.h
 */

#define MAKE_MAGIC(c0,c1,c2,c3,c4,c5,c6,c7) (   \
    (((uint64_t) (uint8_t) (c0)) << (8 * 0)) +  \
    (((uint64_t) (uint8_t) (c1)) << (8 * 1)) +  \
    (((uint64_t) (uint8_t) (c2)) << (8 * 2)) +  \
    (((uint64_t) (uint8_t) (c3)) << (8 * 3)) +  \
    (((uint64_t) (uint8_t) (c4)) << (8 * 4)) +  \
    (((uint64_t) (uint8_t) (c5)) << (8 * 5)) +  \
    (((uint64_t) (uint8_t) (c6)) << (8 * 6)) +  \
    (((uint64_t) (uint8_t) (c7)) << (8 * 7))    \
)

typedef enum _XS_FASTPATH_REQUEST_FORMAT: uint32_t
{
    XS_FASTPATH_JSON_FORMAT = 1,
} XS_FASTPATH_REQUEST_FORMAT;

typedef enum _XS_FASTPATH_REQUEST_METHOD: uint32_t
{
    XS_FASTPATH_OPEN_HANDLE = 1,
    XS_FASTPATH_READ_HANDLE = 2,
    XS_FASTPATH_CLOSE_HANDLE = 3
} XS_FASTPATH_REQUEST_METHOD;

static const uint64_t MAGIC_REQUEST_HEADER = MAKE_MAGIC('H', 's', 'R', 'e', 'q', '0', '0', '1');
static const uint64_t MAGIC_RESPONSE_HEADER = MAKE_MAGIC('H', 's', 'R', 'e', 's', '0', '0', '1');

typedef struct _XS_FASTPATH_HOST_REQUEST_HEADER {
    uint64_t Magic;
    XS_FASTPATH_REQUEST_FORMAT Format;
    XS_FASTPATH_REQUEST_METHOD Method;
    uint64_t Crc64;
    uint64_t UnUsed;
} XS_FASTPATH_HOST_REQUEST_HEADER, *PXS_FASTPATH_HOST_REQUEST_HEADER;

typedef struct _XS_FASTPATH_HOST_RESPONSE_HEADER {
    uint64_t Magic;
    XS_FASTPATH_REQUEST_FORMAT Format;
    uint32_t UnUsed1;
    uint64_t Crc64;
    uint64_t UnUsed2;
} XS_FASTPATH_HOST_RESPONSE_HEADER, *PXS_FASTPATH_HOST_RESPONSE_HEADER;

typedef struct _XS_FASTPATH_REQUEST {
    XS_FASTPATH_HOST_REQUEST_HEADER hdr;
    char body[];
} XS_FASTPATH_REQUEST, *PXS_FASTPATH_REQUEST;

typedef struct _XS_FASTPATH_RESPONSE {
    XS_FASTPATH_HOST_RESPONSE_HEADER hdr;
    char body[];
} XS_FASTPATH_RESPONSE, *PXS_FASTPATH_RESPONSE;

static const int JSON_MAX_RESPONSE_SIZE = (32 * 1024);
static const int EXMSG_MAX_LEN = 128;

static std::map<int, const char *> EXMSGS = {
    {EM_SUCCESS, "Success"},
    {EM_MALLOC_FAILED, "malloc failed"},
    {EM_GSTRUC_FAILED, "GetStringUTFChars Failed"},
    {EM_GPAC_FAILED, "GetPrimitiveArrayCritical Failed"},
    {EM_NOT_PGALIGNED, "Read buffer, or offset, or length not page aligned"},
    {EM_VSC_FAILURE, "VSC Failure"},
    {EM_BAD_JSON, "Bad JSON Response"},
    {EM_DANGLING_RESPONSE, "Dangling registered responses exist"},
    {EM_BAD_LENGTH, "Host service returned zero length response"},
    {EM_NOT_DIRECT_BYTEBUF, "Not a direct-allocated ByteBuffer"},
    {EM_BAD_BUF_LEN, "Bad buffer offset or length"},
    {EM_BAD_BUFFER, "Bad buffer address"},
    {EM_VSC_INIT_FAILED, "VSC Initialization Failed"},
    {EM_BAD_HDR_CRC, "Bad response header CRC"},
    {EM_BAD_HDR_MAGIC, "Bad response header magic"},
    {EM_BAD_JNI_ENV, "Bad JNI Env pointer"},
};

static const std::string fpConnectionExceptionClass = std::string(
        "com/azure/storage/fastpath/exceptions/FastpathConnectionException");
static const std::string fpConnectionExceptionCtorName = std::string(
        "<init>");
static const std::string fpConnectionExceptionSig = "(IILjava/lang/String;)V";

extern SharedLibContext *gSharedLibContext;

JNIEXPORT void JNICALL Java_com_azure_storage_fastpath_driver_FastpathDriver_initContext(JNIEnv *env,
        jobject thisObject,
        jboolean testMode)
{
    UNUSED(thisObject);
    UNUSED(env);

    gSharedLibContext->fastpathJniContext.InitFastpathContext(testMode);
}

static inline void
ReleaseParams (JNIEnv *env,
        jstring clientRequestId,
        const char **pVscClientRequestId,
        const char **pVscInputBuffer,
        char *pVscResponseBuffer)
{
    if (*pVscInputBuffer != NULL) {
        delete[] *pVscInputBuffer;
        *pVscInputBuffer = NULL;
    }

    if (*pVscClientRequestId != NULL) {
        env->ReleaseStringUTFChars(clientRequestId, *pVscClientRequestId);
        *pVscClientRequestId = NULL;
    }

    if (pVscResponseBuffer != NULL) {
        delete[] pVscResponseBuffer;
        pVscResponseBuffer = NULL;
    }
}

static inline int
InitParams (JNIEnv *env,
        jstring clientRequestId,
        XS_FASTPATH_REQUEST_METHOD opType,
        jstring requestParams,
        const char **pVscClientRequestId,
        const char **pVscInputBuffer,
        uint32_t *vscInputBufferLength,
        char **pVscResponseBuffer,
        uint32_t vscResponseLength)
{
    int rc = 0;
    PXS_FASTPATH_REQUEST pRequest = NULL;
    const char *requestBody = NULL;
    size_t requestHeaderLength = 0;
    size_t requestBodyLength = 0;
    Crc64Hash crc64;

    *pVscClientRequestId = NULL;
    *pVscInputBuffer = NULL;
    *pVscResponseBuffer = NULL;

    *pVscClientRequestId = env->GetStringUTFChars(clientRequestId, NULL);
    if (*pVscClientRequestId == NULL) {
        rc = EM_GSTRUC_FAILED;
        goto err_out;
    }

    requestBody = env->GetStringUTFChars(requestParams, NULL);
    if (requestBody == NULL) {
        rc = EM_GSTRUC_FAILED;
        goto err_out;
    }
    requestHeaderLength = sizeof(XS_FASTPATH_REQUEST);
    requestBodyLength = env->GetStringUTFLength(requestParams);
    *vscInputBufferLength = requestHeaderLength + requestBodyLength;
    try {
        *pVscInputBuffer = new char [*vscInputBufferLength]();
        *pVscResponseBuffer = new char[vscResponseLength]();
    } catch (std::bad_alloc& e) {
        rc = EM_MALLOC_FAILED;
        goto err_out;
    }
    pRequest = (PXS_FASTPATH_REQUEST) *pVscInputBuffer;
    pRequest->hdr.Magic = MAGIC_REQUEST_HEADER;
    pRequest->hdr.Format = XS_FASTPATH_JSON_FORMAT;
    pRequest->hdr.Method = opType;
    memcpy(pRequest->body, requestBody, requestBodyLength);
    env->ReleaseStringUTFChars(requestParams, requestBody);

    {
        uint64_t computedCrc = crc64.Compute(
                (const uint8_t *) (*pVscInputBuffer),
                *vscInputBufferLength,
                0);
        pRequest->hdr.Crc64 = computedCrc;
    }

    return rc;

err_out:
    ReleaseParams(env,
        clientRequestId,
        pVscClientRequestId,
        pVscInputBuffer,
        *pVscResponseBuffer);
    return rc;
}

static int IssueVscOp(const char *pVscClientRequestId,
        jint timeoutSec,
        const char *pVscInputBuffer,
        uint32_t vscInputBufferLength,
        signed char *pVscReadBuffer,
        uint32_t vscReadBufferLength,
        char *pVscResponseBuffer,
        int32_t *localErrno,
        uint32_t *hostResponseLength)
{
    int rc = 0;
    uint32_t timeoutMsec = timeoutSec * 1000;
    PXS_FASTPATH_REQUEST pRequest = (PXS_FASTPATH_REQUEST) pVscInputBuffer;

    if (UNLIKELY(gSharedLibContext->fastpathJniContext.testMode == true)) {
        FakeResponse *response = LookupResponse(pVscClientRequestId);
        if (response != NULL) {
            PXS_FASTPATH_HOST_RESPONSE_HEADER pResponseHdr =
                (PXS_FASTPATH_HOST_RESPONSE_HEADER) pVscResponseBuffer;
            memset(pResponseHdr, 0x0, sizeof(XS_FASTPATH_HOST_RESPONSE_HEADER));
            if (!response->badResponseHeader) {
                pResponseHdr->Magic = MAGIC_RESPONSE_HEADER;
            }
            memcpy(pVscResponseBuffer + sizeof(XS_FASTPATH_HOST_RESPONSE_HEADER),
                    response->jsonResponseString,
                    response->responseLength);
            if (response->fakeRemoteBuffer != NULL && pVscReadBuffer != NULL) {
                signed char *fakeData = response->fakeRemoteBuffer +
                    response->remoteFileOffset;
                memcpy(pVscReadBuffer,
                        fakeData,
                        (response->readLength <= vscReadBufferLength) ?
                        response->readLength : vscReadBufferLength);
            }

            Crc64Hash crc64;
            *hostResponseLength = sizeof(XS_FASTPATH_HOST_RESPONSE_HEADER) +
                response->responseLength;
            uint64_t computedCrc64 = crc64.Compute(
                    (const uint8_t *) pVscResponseBuffer,
                    *hostResponseLength,
                    0);
            pResponseHdr->Crc64 = computedCrc64;

            goto out_process_response;
        }
    }

    switch (pRequest->hdr.Method) {
        case XS_FASTPATH_OPEN_HANDLE: {
            UNUSED(pVscReadBuffer);
            UNUSED(vscReadBufferLength);

            rc = VscOpenBlob(pVscClientRequestId,
                timeoutMsec,
                pVscInputBuffer,
                vscInputBufferLength,
                pVscResponseBuffer,
                JSON_MAX_RESPONSE_SIZE,
                localErrno,
                hostResponseLength);
            break;
        }

        case XS_FASTPATH_READ_HANDLE: {
            rc = VscGetBlob(pVscClientRequestId,
                timeoutMsec,
                pVscInputBuffer,
                vscInputBufferLength,
                (char *) pVscReadBuffer,
                vscReadBufferLength,
                pVscResponseBuffer,
                JSON_MAX_RESPONSE_SIZE,
                localErrno,
                hostResponseLength);
            break;
        }

        case XS_FASTPATH_CLOSE_HANDLE: {
            UNUSED(pVscReadBuffer);
            UNUSED(vscReadBufferLength);

            rc = VscCloseBlob(pVscClientRequestId,
                timeoutMsec,
                pVscInputBuffer,
                vscInputBufferLength,
                pVscResponseBuffer,
                JSON_MAX_RESPONSE_SIZE,
                localErrno,
                hostResponseLength);
            break;
        }

        default: {
            rc = -1;
            *localErrno = ENOTSUP;
            break;
        }
    }

out_process_response:
    if (rc == 0) {
        PXS_FASTPATH_RESPONSE pResponse =
            (PXS_FASTPATH_RESPONSE) pVscResponseBuffer;
        if (pResponse->hdr.Magic != MAGIC_RESPONSE_HEADER) {
            rc = EM_BAD_HDR_MAGIC;
            goto err_out;
        }

        if (*hostResponseLength > JSON_MAX_RESPONSE_SIZE ||
                *hostResponseLength < sizeof(XS_FASTPATH_HOST_RESPONSE_HEADER)) {
            rc = EM_BAD_BUF_LEN;
            goto err_out;
        }

        Crc64Hash crc64;
        uint64_t currentCrc64 = pResponse->hdr.Crc64;
        pResponse->hdr.Crc64 = 0;
        uint64_t computedCrc64 = crc64.Compute((const uint8_t *) pResponse,
                                               *hostResponseLength,
                                               0);
        pResponse->hdr.Crc64 = currentCrc64;
        if (pResponse->hdr.Crc64 != computedCrc64) {
            rc = EM_BAD_HDR_CRC;
        }
    }

err_out:
    return rc;
}

jthrowable MakeFastpathConnectionException(JNIEnv *env, const int rc, const int localErrno)
{
    jthrowable fpException = NULL;
    char exceptionMsg[EXMSG_MAX_LEN];
    jclass fpConnectionExceptionClassref = (jclass) env->FindClass(
            fpConnectionExceptionClass.c_str());
    jmethodID fpConnectionExceptionCtor = env->GetMethodID(
            fpConnectionExceptionClassref,
            fpConnectionExceptionCtorName.c_str(),
            fpConnectionExceptionSig.c_str());

    if (rc == 0) {
        goto err_check_exception;
    }

    if ((EM_SUCCESS < rc) && (rc < EM_MAX)) {
        snprintf(exceptionMsg, EXMSG_MAX_LEN, "%s", EXMSGS[rc]);
    } else if (rc < 0) {
        if (localErrno == EAGAIN        ||
                localErrno == EFAULT    ||
                localErrno == EINVAL) {
            snprintf(exceptionMsg, EXMSG_MAX_LEN, "%s", strerror(localErrno));
        } else if (localErrno == ENODEV) {
            snprintf(exceptionMsg, EXMSG_MAX_LEN, "%s", "VSP not present");
        } else if (localErrno == EHOSTDOWN) {
            snprintf(exceptionMsg, EXMSG_MAX_LEN, "%s", "Host Service error");
        } else if (localErrno == EADDRNOTAVAIL) {
            snprintf(exceptionMsg, EXMSG_MAX_LEN, "%s", "VSC not available");
        } else if (localErrno == EILSEQ) {
            snprintf(exceptionMsg, EXMSG_MAX_LEN, "%s",
                    "VSC/HostSrv returned zero-length response");
        } else {
            snprintf(exceptionMsg, EXMSG_MAX_LEN, "%s", strerror(localErrno));
        }
    }

    fpException = (jthrowable) env->NewObject(fpConnectionExceptionClassref,
            fpConnectionExceptionCtor,
            rc,
            localErrno,
            env->NewStringUTF(exceptionMsg));
    if (fpException == NULL) {
        goto err_check_exception;
    }

    return fpException;

err_check_exception:
    if (fpException == NULL && env->ExceptionCheck()) {
        fpException = env->ExceptionOccurred();
    }

    return fpException;
}

JNIEXPORT jstring JNICALL Java_com_azure_storage_fastpath_driver_FastpathDriver_open(JNIEnv *env,
        jobject thisObject,
        jint timeoutSec,
        jstring clientRequestId,
        jstring requestParams)
{
    const char *pVscClientRequestId = NULL;
    const char *pVscInputBuffer = NULL;
    uint32_t vscInputBufferLength = 0;
    char *pVscResponseBuffer = NULL;
    uint32_t hostResponseLength = 0;
    int rc = 0;
    int localErrno = 0;
    jstring jsonResponseBuffer = NULL;

    UNUSED(thisObject);

    rc = InitParams(env,
            clientRequestId,
            XS_FASTPATH_OPEN_HANDLE,
            requestParams,
            &pVscClientRequestId,
            &pVscInputBuffer,
            &vscInputBufferLength,
            &pVscResponseBuffer,
            JSON_MAX_RESPONSE_SIZE);
    if (rc) {
        goto err_out;
    }

    rc = IssueVscOp(pVscClientRequestId,
            timeoutSec,
            pVscInputBuffer,
            vscInputBufferLength,
            NULL,
            0,
            pVscResponseBuffer,
            &localErrno,
            &hostResponseLength);
    if (rc == 0) {
        std::string responseBody(((PXS_FASTPATH_RESPONSE) pVscResponseBuffer)->body,
                hostResponseLength - sizeof(XS_FASTPATH_HOST_RESPONSE_HEADER));
        jsonResponseBuffer = env->NewStringUTF(responseBody.c_str());
        if (jsonResponseBuffer == NULL) {
            rc = -1;
            if (env->ExceptionCheck()) {
                localErrno = ENOMEM;
                env->ExceptionDescribe();
                env->ExceptionClear();
            }
            goto err_out;
        }
    }

err_out:
    ReleaseParams(env,
            clientRequestId,
            &pVscClientRequestId,
            &pVscInputBuffer,
            pVscResponseBuffer);
    if (rc != 0) {
        jthrowable fpEx = env->ExceptionCheck() ?
            env->ExceptionOccurred() :
            MakeFastpathConnectionException(env, rc, localErrno);
        env->Throw(fpEx);
    }
    return jsonResponseBuffer;
}

JNIEXPORT jstring JNICALL Java_com_azure_storage_fastpath_driver_FastpathDriver_readByteArray(
        JNIEnv *env,
        jobject thisObject,
        jint timeoutSec,
        jstring clientRequestId,
        jstring requestParams,
        jint readBufferOffset,
        jint readLength,
        jbyteArray readByteArray)
{
    jboolean isCopy = JNI_FALSE;
    const char *pVscClientRequestId = NULL;
    const char *pVscInputBuffer = NULL;
    uint32_t vscInputBufferLength = 0;
    signed char *readBuffer = NULL;
    signed char *pVscReadBuffer = NULL;
    uint32_t vscReadBufferLength = 0;
    char *pVscResponseBuffer = NULL;
    uint32_t hostResponseLength = 0;

    int rc = 0;
    int32_t localErrno = 0;
    jstring jsonResponseBuffer = NULL;

    uint32_t readBufferSize = 0;

    UNUSED(thisObject);
    UNUSED(vscInputBufferLength);

    rc = InitParams(env,
            clientRequestId,
            XS_FASTPATH_READ_HANDLE,
            requestParams,
            &pVscClientRequestId,
            &pVscInputBuffer,
            &vscInputBufferLength,
            &pVscResponseBuffer,
            JSON_MAX_RESPONSE_SIZE);
    if (rc) {
        goto err_out;
    }

    readBufferSize = env->GetArrayLength(readByteArray);

    readBuffer = env->GetByteArrayElements(readByteArray, &isCopy);
    if (readBuffer == NULL) {
        rc = EM_GPAC_FAILED;
        goto err_out;
    }
    pVscReadBuffer = readBuffer + readBufferOffset;
    vscReadBufferLength =
        ((uint32_t)(readBufferOffset + readLength) <= readBufferSize) ?
        readLength :
        (readBufferSize - readBufferOffset);

    rc = IssueVscOp(pVscClientRequestId,
            timeoutSec,
            pVscInputBuffer,
            vscInputBufferLength,
            pVscReadBuffer,
            vscReadBufferLength,
            pVscResponseBuffer,
            &localErrno,
            &hostResponseLength);
    env->ReleaseByteArrayElements(readByteArray, readBuffer, 0);
    if (rc == 0) {
        std::string responseBody(((PXS_FASTPATH_RESPONSE) pVscResponseBuffer)->body,
                hostResponseLength - sizeof(XS_FASTPATH_HOST_RESPONSE_HEADER));
        jsonResponseBuffer = env->NewStringUTF(responseBody.c_str());
        if (jsonResponseBuffer == NULL) {
            rc = -ECANCELED;
            if (env->ExceptionCheck()) {
                localErrno = ENOMEM;
                env->ExceptionDescribe();
                env->ExceptionClear();
            }
            goto err_out;
        }
    }

err_out:
    ReleaseParams(env,
        clientRequestId,
        &pVscClientRequestId,
        &pVscInputBuffer,
        pVscResponseBuffer);
    if (rc != 0) {
        jthrowable fpEx = env->ExceptionCheck() ?
            env->ExceptionOccurred() :
            MakeFastpathConnectionException(env, rc, localErrno);
        env->Throw(fpEx);
    }

    return jsonResponseBuffer;
}

JNIEXPORT jstring JNICALL Java_com_azure_storage_fastpath_driver_FastpathDriver_read(JNIEnv *env,
        jobject thisObject,
        jint timeoutSec,
        jstring clientRequestId,
        jstring requestParams,
        jint readBufferOffset,
        jint readLength,
        jobject readBuffer)
{
    const char *pVscClientRequestId = NULL;
    const char *pVscInputBuffer = NULL;
    uint32_t vscInputBufferLength = 0;
    signed char *pVscReadBuffer = NULL;
    uint32_t vscReadBufferLength = 0;
    int32_t readBufferSize = 0;
    char *pVscResponseBuffer = NULL;
    uint32_t hostResponseLength = 0;

    signed char *pReadBuffer = NULL;
    int rc = 0;
    int localErrno = 0;
    jstring jsonResponseBuffer = NULL;


    UNUSED(thisObject);

    if (readBufferOffset < 0 || readLength < 0) {
        rc = EM_BAD_BUF_LEN;
        localErrno = EINVAL;
        goto err_out;
    }

    rc = InitParams(env,
            clientRequestId,
            XS_FASTPATH_READ_HANDLE,
            requestParams,
            &pVscClientRequestId,
            &pVscInputBuffer,
            &vscInputBufferLength,
            &pVscResponseBuffer,
            JSON_MAX_RESPONSE_SIZE);
    if (rc) {
        goto err_out;
    }

    pReadBuffer = (signed char *) env->GetDirectBufferAddress(readBuffer);
    if (pReadBuffer == NULL) {
        localErrno = EINVAL;
        rc = EM_NOT_DIRECT_BYTEBUF;
        goto err_out;
    }
    readBufferSize = env->GetDirectBufferCapacity(readBuffer);
    if (readBufferSize == -1) {
        localErrno = EINVAL;
        rc = EM_NOT_DIRECT_BYTEBUF;
        goto err_out;
    }

    pVscReadBuffer = pReadBuffer + readBufferOffset;
    if ((readBufferOffset + readLength) > readBufferSize) {
        rc = EM_BAD_BUF_LEN;
        localErrno = EINVAL;
        goto err_out;
    }
    vscReadBufferLength = readLength;

    rc = IssueVscOp(pVscClientRequestId,
            timeoutSec,
            pVscInputBuffer,
            vscInputBufferLength,
            pVscReadBuffer,
            vscReadBufferLength,
            pVscResponseBuffer,
            &localErrno,
            &hostResponseLength);
    if (rc == 0) {
        std::string responseBody(((PXS_FASTPATH_RESPONSE) pVscResponseBuffer)->body,
                hostResponseLength - sizeof(XS_FASTPATH_HOST_RESPONSE_HEADER));
        jsonResponseBuffer = env->NewStringUTF(responseBody.c_str());
        if (jsonResponseBuffer == NULL) {
            rc = -1;
            if (env->ExceptionCheck()) {
                localErrno = ENOMEM;
                env->ExceptionDescribe();
                env->ExceptionClear();
            }
            goto err_out;
        }
    }

err_out:
    ReleaseParams(env,
        clientRequestId,
        &pVscClientRequestId,
        &pVscInputBuffer,
        pVscResponseBuffer);
    if (rc != 0) {
        jthrowable fpEx = env->ExceptionCheck() ?
            env->ExceptionOccurred() :
            MakeFastpathConnectionException(env, rc, localErrno);
        env->Throw(fpEx);
    }

    return jsonResponseBuffer;
}

JNIEXPORT jstring JNICALL Java_com_azure_storage_fastpath_driver_FastpathDriver_close(JNIEnv *env,
        jobject thisObject,
        jint timeoutSec,
        jstring clientRequestId,
        jstring requestParams)
{
    const char *pVscClientRequestId = NULL;
    const char *pVscInputBuffer = NULL;
    uint32_t vscInputBufferLength = 0;
    char *pVscResponseBuffer = NULL;
    uint32_t hostResponseLength = 0;

    int rc = 0;
    int localErrno = 0;
    jstring jsonResponseBuffer = NULL;

    UNUSED(thisObject);

    rc = InitParams(env,
            clientRequestId,
            XS_FASTPATH_CLOSE_HANDLE,
            requestParams,
            &pVscClientRequestId,
            &pVscInputBuffer,
            &vscInputBufferLength,
            &pVscResponseBuffer,
            JSON_MAX_RESPONSE_SIZE);
    if (rc) {
        goto err_out;
    }

    rc = IssueVscOp(pVscClientRequestId,
            timeoutSec,
            pVscInputBuffer,
            vscInputBufferLength,
            NULL,
            0,
            pVscResponseBuffer,
            &localErrno,
            &hostResponseLength);
    if (rc == 0) {
        std::string responseBody(((PXS_FASTPATH_RESPONSE) pVscResponseBuffer)->body,
                hostResponseLength - sizeof(XS_FASTPATH_HOST_RESPONSE_HEADER));
        jsonResponseBuffer = env->NewStringUTF(responseBody.c_str());
        if (jsonResponseBuffer == NULL) {
            rc = -1;
            if (env->ExceptionCheck()) {
                localErrno = ENOMEM;
                env->ExceptionDescribe();
                env->ExceptionClear();
            }
            goto err_out;
        }
    }

err_out:
    ReleaseParams(env,
        clientRequestId,
        &pVscClientRequestId,
        &pVscInputBuffer,
        pVscResponseBuffer);
    if (rc != 0) {
        jthrowable fpEx = env->ExceptionCheck() ?
            env->ExceptionOccurred() :
            MakeFastpathConnectionException(env, rc, localErrno);
        env->Throw(fpEx);
    }

    return jsonResponseBuffer;
}
