/*
 * Copyright (c) Microsoft Corporation. All rights reserved.
 * Licensed under the MIT License.
 * See License.txt in the project root for license information.
 */

#include <assert.h>
#include <stdbool.h>
#include <inttypes.h>
#include <pthread.h>

#include <string.h>
#include <stdlib.h>

#include <fpcommon.h>
#include <jni/exceptions.h>
#include "fake_responses.h"

FakeResponse::FakeResponse()
    :clientRequestIdRef {NULL},
    jsonResponseStringRef {NULL},
    clientRequestId {NULL},
    jsonResponseString {NULL},
    responseLength {0},
    readLength {0},
    remoteFileOffset {0},
    fakeRemoteBuffer {NULL},
    badResponseHeader {false}
{
}

FakeResponse::~FakeResponse()
{
}

int FakeResponse::Init(JNIEnv *env,
    jstring fClientRequestId,
    jstring fJsonResponseString,
    jint fReadLength,
    jlong fRemoteFileOffset,
    jobject fRemoteBuffer)
{
    int rc = 0;

    badResponseHeader = false;
    clientRequestIdRef = (jstring) env->NewGlobalRef(fClientRequestId);
    jsonResponseStringRef = (jstring) env->NewGlobalRef(fJsonResponseString);

    clientRequestId = env->GetStringUTFChars(fClientRequestId, NULL);
    if (clientRequestId == NULL) {
        rc = EM_GSTRUC_FAILED;
        goto out;
    }

    jsonResponseString = env->GetStringUTFChars(fJsonResponseString, NULL);
    if (jsonResponseString == NULL) {
        rc = EM_GSTRUC_FAILED;
        goto out;
    }

    responseLength = env->GetStringUTFLength(fJsonResponseString);
    if (responseLength == 0) {
        rc = EM_BAD_LENGTH;
        goto out;
    }

    if (fReadLength > 0 && fRemoteBuffer != NULL) {
        int fRemoteBufLen = env->GetDirectBufferCapacity(fRemoteBuffer);
        readLength = (fReadLength <= (fRemoteBufLen - fRemoteFileOffset)) ?
            fReadLength : (fRemoteBufLen - fRemoteFileOffset);
        remoteFileOffset = fRemoteFileOffset;
        fakeRemoteBuffer = (int8_t *) env->GetDirectBufferAddress(fRemoteBuffer);
        if (fakeRemoteBuffer == NULL) {
            rc = EM_GPAC_FAILED;
            goto out;
        }
    }

out:
    return rc;
}


void FakeResponse::CleanupJniRefs(JNIEnv *env)
{
    if (jsonResponseString != NULL) {
        env->ReleaseStringUTFChars(jsonResponseStringRef,
                jsonResponseString);
        env->DeleteGlobalRef(jsonResponseStringRef);
    }

    if (clientRequestId != NULL) {
        env->ReleaseStringUTFChars(clientRequestIdRef,
                clientRequestId);
        env->DeleteGlobalRef(clientRequestIdRef);
    }

    jsonResponseString = NULL;
    clientRequestId = NULL;
}

FakeResponseContext::FakeResponseContext()
{
}

FakeResponseContext::~FakeResponseContext()
{
}

void FakeResponseContext::RegisterResponse(JNIEnv *env,
    jstring fClientRequestId,
    jstring fJsonResponseString,
    jint fReadLength,
    jlong fRemoteFileOffset,
    jobject fRemoteBuffer)
{
    FakeResponse *response = NULL;
    int rc = 0;

    try {
        response = new FakeResponse;
    } catch (std::bad_alloc& e) {
        rc = EM_MALLOC_FAILED;
        goto err_out;
    }

    rc = response->Init(env,
                fClientRequestId,
                fJsonResponseString,
                fReadLength,
                fRemoteFileOffset,
                fRemoteBuffer);
    if (rc != 0) {
        goto err_out;
    }

    {
        std::lock_guard<std::mutex> lk_guard {lock};

        if (rmap.find(response->clientRequestId) == rmap.end()) {
            rmap[response->clientRequestId] = response;
        } else {
            if ((rmap[response->clientRequestId])->badResponseHeader) {
                response->badResponseHeader = true;
                rmap[response->clientRequestId] = response;
            }
        }
    }

    return;

err_out:
    if (response != NULL) {
        response->CleanupJniRefs(env);
        delete response;
        response = NULL;
    }

    if (rc != 0) {
        jthrowable fpEx = MakeFastpathConnectionException(env, rc, errno);
        env->Throw(fpEx);
    }

}

void FakeResponseContext::UnregisterResponse(JNIEnv *env, jstring fClientRequestId)
{
    FakeResponse *response = NULL;
    const char *key = NULL;
    int rc = 0;

    key = env->GetStringUTFChars(fClientRequestId, NULL);
    if (key == NULL) {
        rc = EM_GSTRUC_FAILED;
        goto out_throw;
    }

    {
        std::lock_guard<std::mutex> lk_guard {lock};
        auto lookup = rmap.find(key);
        if (lookup != rmap.end()) {
            response = lookup->second;
            rmap.erase(key);
        }
    }

    if (response != NULL) {
        if (!env->IsSameObject(fClientRequestId, response->clientRequestIdRef)) {
            assert(0);
        }
        env->ReleaseStringUTFChars(response->clientRequestIdRef, key);
        response->CleanupJniRefs(env);
        delete response;
    }

    return;

out_throw:
    if (rc != 0) {
        jthrowable fpEx = MakeFastpathConnectionException(env, rc, errno);
        env->Throw(fpEx);
    }
}

void FakeResponseContext::SetBadResponseHeader(JNIEnv *env, jstring fClientRequestId)
{
    int rc = 0;
    FakeResponse *response = NULL;

    try {
        response = new FakeResponse;
    } catch (std::bad_alloc& e) {
        rc = EM_MALLOC_FAILED;
        goto err_out;
    }

    response->badResponseHeader = true;
    response->clientRequestId = env->GetStringUTFChars(fClientRequestId, NULL);

    {
        std::lock_guard<std::mutex> lk_guard {lock};
        if (rmap.find(response->clientRequestId) == rmap.end()) {
            response->clientRequestIdRef = (jstring) env->NewGlobalRef(fClientRequestId);
            rmap[response->clientRequestId] = response;
        } else {
            (rmap[response->clientRequestId])->badResponseHeader = true;
        }
    }

    return;

err_out:
    if (response != NULL) {
        delete response;
    }

    if (rc != 0) {
        jthrowable fpEx = MakeFastpathConnectionException(env, rc, errno);
        env->Throw(fpEx);
    }
}

FakeResponse * FakeResponseContext::LookupResponse(const char *fClientRequestId)
{
    FakeResponse *response = NULL;

    std::lock_guard<std::mutex> lk_guard {lock};
    auto lookup = rmap.find(fClientRequestId);
    if (lookup != rmap.end()) {
        response = lookup->second;
    }

    return response;
}
