#include <cerrno>
#include <cstddef>
#include <cstring>

#include <mutex>

#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include <fpcommon.h>

#include "fastpath.h"

using stdclk = std::chrono::steady_clock;

static constexpr uint32_t VSC_INVALID_RESPONSE_STATUS = 0X0ACE0ACE;
static constexpr uint32_t HRESULT_W32_TIMEOUT = 0x80070102;
static constexpr std::chrono::seconds VSC_REOPEN_BACKOFF_SEC {60};
static constexpr uint64_t VSC_CTX_HDR_MAGIC = 0x0A0A0A0A0BADCAFE;
static constexpr uint64_t VSC_CTX_FTR_MAGIC = 0x0C0C0C0C0BADCAFE;
static constexpr const char *VSC_DEV_PATH = "/dev/azure_xs_fastpath";

typedef uint32_t __u32;
typedef uint64_t __attribute__((aligned(8))) __aligned_u64;
typedef struct {
    uint8_t b[16];
} guid_t;

#ifndef FP_MOCKTEST_OS
#include <misc/rimbaud.h>
#else
#include "rimbaud.h"
#endif

struct VscSyncRequest
{
    struct xs_fastpath_request_sync xsFpSyncRequest;

    VscSyncRequest(const char *client_request_id,
                             uint32_t timeout_msec,
                             const char *input_buf,
                             uint32_t input_buf_len,
                             char *data_buf,
                             uint32_t data_buf_len,
                             uint32_t data_valid,
                             char *response_buf,
                             uint32_t response_buf_len)
    {
        memcpy(xsFpSyncRequest.guid.b, client_request_id, sizeof(guid_t));
        xsFpSyncRequest.timeout = timeout_msec;
        xsFpSyncRequest.request_len = input_buf_len;
        xsFpSyncRequest.response_len = response_buf_len;
        xsFpSyncRequest.data_len = data_buf_len;
        xsFpSyncRequest.data_valid = data_valid;
        xsFpSyncRequest.request_buffer = (__aligned_u64) input_buf;
        xsFpSyncRequest.response_buffer = (__aligned_u64) response_buf;
        xsFpSyncRequest.data_buffer = (__aligned_u64) data_buf;

        xsFpSyncRequest.response = {
            .status = VSC_INVALID_RESPONSE_STATUS,
            .response_len = 0,
        };
    }
};

static constexpr uint64_t VSC_IOCTL_RQ = _IOWR('R', 10, xs_fastpath_request_sync);

VscContext::VscContext ()
    :hdrMagic {VSC_CTX_HDR_MAGIC},
     status {VSC_STATUS_NOTAVAIL},
     fdOpenCount {0},
     vscFd {-1},
     vscBadFd {-1},
     vscUnavailTimestamp {stdclk::time_point::max()},
     ftrMagic {VSC_CTX_FTR_MAGIC}
{
}

int VscContext::InitVscDevice ()
{
    int localErrno = 0;

    std::lock_guard<std::mutex> lkGuard {vscCtxLock};

    if (status == VSC_STATUS_AVAIL) {
        goto out;
    }

    fdOpenCount++;
    vscBadFd = vscFd;

    vscFd = open(VSC_DEV_PATH, O_RDWR);
    if (vscFd < 0) {
        localErrno = errno;
        vscUnavailTimestamp = stdclk::now();
        goto out;
    }

    status = VSC_STATUS_AVAIL;

out:
    return localErrno;
}

void VscContext::CleanupVscDevice (bool reinit = false)
{
    std::lock_guard<std::mutex> lkGuard {vscCtxLock};

    if (vscBadFd >= 0) {
        close(vscBadFd);
        vscBadFd = -1;
    }

    if (status != VSC_STATUS_AVAIL ||
            (status == VSC_STATUS_AVAIL && reinit == true)) {
        goto out;
    }

    if (vscFd >= 0) {
        close(vscFd);
        vscFd = -1;
    }

    if (status != VSC_STATUS_NOTAVAIL) {
        status = VSC_STATUS_NOTAVAIL;
    }

out:
    return;
}

VscContext::~VscContext ()
{
    CleanupVscDevice();
}

int VscContext::ReinitVscFd ()
{
    int localErrno = 0;

    if (status == VSC_STATUS_NOTAVAIL) {
        stdclk::time_point now = stdclk::now();
        std::chrono::seconds elapsedTimeSec =
            std::chrono::duration_cast<std::chrono::seconds>(
                    now - vscUnavailTimestamp);

        if (elapsedTimeSec < VSC_REOPEN_BACKOFF_SEC) {
            localErrno = EADDRNOTAVAIL;
            goto out;
        }

        bool reinit = true;
        localErrno = InitVscDevice();
        CleanupVscDevice(reinit);
    }

out:
    return localErrno;
}

int VscContext::IssueVscSyncRequest (IN const char *clientRequestId,
    IN uint32_t timeoutMsec,
    IN const char *pInputBuf,
    IN uint32_t inputBufLen,
    OUT char *pDataBuf,
    IN uint32_t dataBufLen,
    IN uint32_t dataValid,
    OUT char *pResponseBuf,
    IN uint32_t responseBufLen,
    OUT int32_t *pLocalErrno,
    OUT uint32_t *pHostResponseLen)
{
    int rc = 0;
    VscSyncRequest xsSyncRequest (clientRequestId,
            timeoutMsec,
            pInputBuf,
            inputBufLen,
            pDataBuf,
            dataBufLen,
            dataValid,
            pResponseBuf,
            responseBufLen);

    *pLocalErrno = 0;
    if (UNLIKELY(status != VSC_STATUS_AVAIL)) {
        *pLocalErrno = ReinitVscFd();
        if (*pLocalErrno != 0) {
            rc = -1;
            goto err_out;
        }
    }

    rc = ioctl(vscFd, VSC_IOCTL_RQ, &(xsSyncRequest.xsFpSyncRequest));
    if (rc != 0)  {
        if (errno == ENODEV) {
            vscUnavailTimestamp = stdclk::now();
            status = VSC_STATUS_NOTAVAIL;
        }
        *pLocalErrno = errno;
    }

    if (xsSyncRequest.xsFpSyncRequest.response.status == HRESULT_W32_TIMEOUT) {
        rc = -1;
        *pLocalErrno = EHOSTDOWN;
    }

    if (xsSyncRequest.xsFpSyncRequest.response.response_len == 0) {
        if (rc == 0) {
            rc = -1;
        }
        *pLocalErrno = EILSEQ;
    }

    if (pHostResponseLen != NULL) {
        *pHostResponseLen = xsSyncRequest.xsFpSyncRequest.response.response_len;
    }

err_out:
    return rc;
}
