/*
 * Copyright (c) Microsoft Corporation. All rights reserved.
 * Licensed under the MIT License.
 * See License.txt in the project root for license information.
 */

package com.azure.storage.fastpath;

import java.nio.ByteBuffer;
import java.util.Random;
import java.util.Arrays;
import java.util.regex.Pattern;
import java.util.regex.Matcher;

import com.azure.storage.fastpath.driver.FastpathDriver;
import com.azure.storage.fastpath.exceptions.FastpathConnectionException;
import com.azure.storage.fastpath.utility.SharedLibHelper;

public class FPTest {
    static {
        System.loadLibrary("fastpath-" + SharedLibHelper.getAndUpdateOSVersionCPU());
    }

    static void fillRandomBytes (ByteBuffer bbuf, int length)
    {
        int rlen = 0;
        if (length == 0) {
            rlen = bbuf.remaining();
        } else {
            rlen = (length < bbuf.remaining()) ? length : bbuf.remaining();
        }

        Random rnd = new Random();
        byte rData[] = new byte[rlen];

        rnd.nextBytes(rData);
        bbuf.put(rData);

        if (!bbuf.hasRemaining()) {
            bbuf.rewind();
        }
    }

    static void fillRepeatedPattern (ByteBuffer bbuf, byte [] pattern)
    {
        while (bbuf.hasRemaining()) {
            int plen = (pattern.length < bbuf.remaining()) ?
                pattern.length : bbuf.remaining();
            bbuf.put(pattern, 0, plen);
        }
        bbuf.rewind();
    }

    static void fillIntIndices (ByteBuffer bbuf) {
        int index = 0;
        int mask = 0xc3c30000;
        while (bbuf.hasRemaining()) {
            bbuf.putInt(mask | index);
            index++;
        }
        bbuf.rewind();
    }

    static void dumpByteBuffer (String tag,
            ByteBuffer bbuf,
            int offset,
            int length)
    {
        int extract_len = (length == 0) ? bbuf.capacity() : length;
        byte[] b_array = new byte[extract_len];
        int saved_position = bbuf.position();
        bbuf.rewind();
        bbuf.position(offset);
        bbuf.get(b_array, 0, extract_len);
        bbuf.position(saved_position);

        System.out.printf("%s: ", tag);
        for (byte b: b_array) {
            System.out.printf("0x%x ", b);
        }
        System.out.println();
    }

    static boolean compareByteBuffers (int len,
            ByteBuffer buf1,
            int off1,
            ByteBuffer buf2,
            int off2)
    {
        byte[] buf1_ex = new byte[len];
        byte[] buf2_ex = new byte[len];

        int saved_pos1 = buf1.position();
        int saved_pos2 = buf2.position();

        buf1.rewind();
        buf1.position(off1);
        buf1.get(buf1_ex);

        buf2.rewind();
        buf2.position(off2);
        buf2.get(buf2_ex);

        boolean ret = Arrays.equals(buf1_ex, buf2_ex);

        buf1.position(saved_pos1);
        buf2.position(saved_pos2);

        return ret;
    }

    static void testOpenVSC()
    {
        String openResponse = "Before Open";
        int timeoutMsec = 3000;
        String clientRequestID = "clientRequestID_1001";
        String slReqParams = "{\"foo\" : \"42 - 1\"}";

        Pattern patClientReqId = Pattern.compile(
                "\"clientRequestIdentifier\" : \"defaultClientRequestId\"");
        try {
            FastpathDriver fpdriver = new FastpathDriver();
            fpdriver.initContext(false);
            openResponse = fpdriver.open(timeoutMsec,
                    clientRequestID,
                    slReqParams);
        } catch (FastpathConnectionException ex) {
            System.out.println("UT: FAILED: testOpenVSC");
            System.out.println("======  caught exception ============");
            System.out.println(ex.toString());
            System.out.println("=====================================\n");
            return;
        }

        Matcher openResponseMatcher = patClientReqId.matcher(openResponse);
        if (openResponseMatcher.find()) {
            System.out.println("UT: PASSED: testOpenVSC");
        } else {
            System.out.println("UT: FAILED: testOpenVSC ");
        }
    }

    static void testOpenFakeUnregistered()
    {
        int timeoutMsec = 3000;

        String clientRequestID = "clientRequestID_1001";
        String slReqParams = "{\"foo\" : \"42 - 1\"}";
        String fakeOpenResponse = "{\"op\" : \"open\", \"type\": \"FAKE\"}";

        String openResponse;

        Pattern patClientReqId = Pattern.compile(
                "\"clientRequestIdentifier\" : \"defaultClientRequestId\"");



        try {
            FastpathDriver fpdriver = new FastpathDriver();
            fpdriver.initContext(true);
            fpdriver.registerOpenResponse(clientRequestID, fakeOpenResponse);
            fpdriver.unregisterOpenResponse(clientRequestID, fakeOpenResponse);

            openResponse = fpdriver.open(timeoutMsec,
                    clientRequestID,
                    slReqParams);
        } catch (FastpathConnectionException ex) {
            System.out.println("UT: FAILED: testOpenFakeUnregistered");
            System.out.println("======  caught exception ============");
            System.out.println(ex.toString());
            System.out.println("=====================================\n");
            return;
        }

        Matcher openResponseMatcher = patClientReqId.matcher(openResponse);
        if (openResponseMatcher.find()) {
            System.out.println("UT: PASSED: testOpenFakeUnregistered");
        } else {
            System.out.println("UT: FAILED: testOpenFakeUnregistered");
        }
    }

    static void testReadVSC()
    {
        String rdResponse = "Before Read";
        int timeoutMsec = 3000;
        String clientRequestID = "clientRequestID_1002";
        String slReqParams = "{\"bar\" : \"42 + 1\"}";

        int readLength = 4;
        int rdBufOffset = 0;
        long remoteFileOffset = 4;
        int rdBufCapacity = 8192;
        ByteBuffer rdBuf = ByteBuffer.allocateDirect(rdBufCapacity);
        rdBuf.clear();
        fillIntIndices(rdBuf);

        try {
            FastpathDriver fpdriver = new FastpathDriver();
            fpdriver.initContext(false);
            rdResponse = fpdriver.read(timeoutMsec,
                clientRequestID,
                slReqParams,
                rdBufOffset,
                readLength,
                rdBuf);
        } catch (Exception ex) {
            System.out.println("UT: FAILED: testReadVSC");
            System.out.println("======  caught exception ============");
            System.out.println(ex.toString());
            System.out.println("=====================================\n");
            return;
        }
    }

    static void testCloseVSC()
    {
        String closeResponse = "Before close";
        int timeoutMsec = 3000;
        String clientRequestID = "clientRequestID_1001";
        String slReqParams = "{\"foo\" : \"42 - 1\"}";

        Pattern patClientReqId = Pattern.compile(
                "\"clientRequestIdentifier\" : \"defaultClientRequestId\"");

        try {
            FastpathDriver fpdriver = new FastpathDriver();
            fpdriver.initContext(false);
            closeResponse = fpdriver.close(timeoutMsec,
                    clientRequestID,
                    slReqParams);
        } catch (Exception ex) {
            System.out.println("UT: FAILED: testCloseVSC");
            System.out.println("======  caught exception ============");
            System.out.println(ex.toString());
            System.out.println("=====================================\n");
            return;
        }

        Matcher closeResponseMatcher = patClientReqId.matcher(closeResponse);
        if (closeResponseMatcher.find()) {
            System.out.println("UT: PASSED: testCloseVSC");
        } else {
            System.out.println("UT: FAILED: testCloseVSC");
        }
    }

    static void testFakeResponses()
    {
        int timeoutMsec = 3000;
        String openCreqID = "clientReqID_1001";
        String openReqParams = "{\"foo\" : \"42 - 1\"}";
        String openFakeResponse = "{\"op\" : \"open\", \"type\": \"FAKE\"}";
        String openResponse;

        String readCreqID = "clientRequestID_1002";
        String readReqParams = "Some Random parameters I have no idea about";
        String readFakeResponse = "{"   +
            "\"op\": \"read\","         +
            "\"length\": \"16384\","    +
            "\"type\": \"FAKE!!!\""     +
        "}";
        int readLength = 16;
        int rdBufOffset = 4;
        int remoteFileOffset = 8;
        int fakeRdBufCapacity = 8192;
        int rdBufCapacity = 8192;
        assert(rdBufOffset + readLength < fakeRdBufCapacity);
        ByteBuffer fakeRdBuffer =
            ByteBuffer.allocateDirect(fakeRdBufCapacity);
        fakeRdBuffer.clear();
        fillRandomBytes(fakeRdBuffer, 0);
        fakeRdBuffer.rewind();

        ByteBuffer rdBuffer = ByteBuffer.allocateDirect(rdBufCapacity);
        rdBuffer.clear();
        byte[] pattern1 = { 0x2D, 0x2D, 0x2D, 0x2D };
        fillRepeatedPattern(rdBuffer, pattern1);
        String readResponse;

        String closeCreqID = "clientReqID_1003";
        String closeReqParams = "{\"foo\" : \"42 - 1\"}";
        String closeFakeResponse = "{\"op\" : \"close\", \"type\": \"FAKE\"}";
        String closeResponse;


        try {
            FastpathDriver fpd = new FastpathDriver();
            fpd.initContext(true);
            fpd.registerOpenResponse(openCreqID, openFakeResponse);
            fpd.registerReadResponse(readCreqID,
                    readFakeResponse,
                    readLength,
                    remoteFileOffset,
                    fakeRdBuffer);
            fpd.registerCloseResponse(closeCreqID, closeFakeResponse);

            openResponse = fpd.open(timeoutMsec, openCreqID, openReqParams);
            if (openResponse.equals(openFakeResponse)) {
                System.out.println("UT: PASSED: testOpenFakeResponse");
            } else {
                System.out.println("UT: FAILED: OpenFakeResponse");
            }
            readResponse = fpd.read(timeoutMsec,
                    readCreqID,
                    readReqParams,
                    rdBufOffset,
                    readLength,
                    rdBuffer);
            boolean readCompareResult = compareByteBuffers(readLength,
                fakeRdBuffer,
                remoteFileOffset,
                rdBuffer,
                rdBufOffset);
            if (readCompareResult && readResponse.equals(readFakeResponse)) {
                System.out.println("UT: PASSED: testReadFakeResponse");
            } else {
                System.out.println("UT: FAILED: testReadFakeResponse");
            }

            closeResponse = fpd.close(timeoutMsec, closeCreqID, closeReqParams);
            if (closeResponse.equals(closeFakeResponse)) {
                System.out.println("UT: PASSED: testCloseFakeResponse");
            } else {
                System.out.println("UT: FAILED: testCloseFakeResponse");
            }

            fpd.registerBadResponseHdr(openCreqID);
            try {
                openResponse = fpd.open(timeoutMsec, openCreqID, openReqParams);
            } catch (FastpathConnectionException ex) {
                Pattern patBadResponseHdr = Pattern.compile(
                        "Bad response header magic");
                Matcher badResponseMatcher =
                    patBadResponseHdr.matcher(ex.toString());
                if ((ex.getErrNum() == 0) &&
                    (ex.getRetCode() == 14) && (badResponseMatcher.find())) {
                        System.out.println("UT: PASSED: testBadResponseHdr");
                } else {
                    System.out.println("UT: FAILED: testBadResponseHdr");
                }
            }
        } catch (Exception ex) {
            System.out.println("testFakeResponses FAILED: ");
            System.out.println(ex.toString());
        }
        System.out.println();
    }

    static void testOpen()
    {
        testOpenVSC();
        testOpenFakeUnregistered();
    }

    static void testRead()
    {
        testReadVSC();
    }

    static void testClose()
    {
        testCloseVSC();
    }


    public static void main(String args[])
        throws InterruptedException
    {
        testFakeResponses();
    }
}
